Strict

Rem
bbdoc: Thunder.ObjectEx
EndRem
Module Thunder.ObjectEx

ModuleInfo "Version: 0.5"
ModuleInfo "Author: Thunder"
ModuleInfo "License: WTFPL"

Import brl.map
Import "copytype.c"

Extern
	Function ParseClass$z(o:ObjectEx, clas:Int Ptr, cb(x:ObjectEx, id, name:Byte Ptr, typ:Byte Ptr, offset))
	Function ParseObject$z(o:ObjectEx, inst:Object, cb(x:ObjectEx, id, name:Byte Ptr, typ:Byte Ptr, offset))

	Rem
	bbdoc: Copies an Object of any type
	returns: Flat Copy of @x
	about:
	Copy an Object of any type.
	This will copy references as they
	are. So references in the copied
	instance will point at the same
	structures as references from the
	source instance.
	EndRem
	Function CopyObject:Object(x:Object)
	
	Rem
	bbdoc: Copies an Object of any type and its children
	returns: Recursive Copy of @x
	about:
	This function copies a whole instance recursively and
	tries to avoid copying child objects more than
	once.
	It parses the @debug_scope which is a
	structure that saves all names of
	fields, methods etc., their types and
	offset.
	Pointers are treated like primitives.
	Arrays they might point to are not copied. 
	EndRem
	Function CopyObjectTree:Object(x:Object)
	
	Rem
	bbdoc: Converts any Object type to Byte Ptr (doing this implicitly in BlitzMax gives you the same pointer + 8 byte)
	returns: Pointer to @x
	EndRem
	Function ObjectToPtr:Byte Ptr(x:Object) = "ObjectPtr"
	
	Rem
	bbdoc: Converts Byte Ptr to Object (impossible in BlitzMax)
	returns: Object from @x
	EndRem
	Function PtrToObject:Object(x:Byte Ptr) = "ObjectPtr"
	
	Rem
	bbdoc: Converts any Object type to Int
	returns: @x as Integer
	EndRem
	Function ObjectToInt:Int(x:Object) = "ObjectPtr"
	
	Rem
	bbdoc: Converts Int to Object
	returns: @x as Object
	EndRem
	Function IntToObject:Object(x:Int) = "ObjectPtr"

EndExtern

Function oex_cb(o:ObjectEx, id, name:Byte Ptr, typ:Byte Ptr, offset)
	o.new_ident id, String.FromCString(name), String.FromCString(typ), offset
EndFunction

Rem
bbdoc: Get Pointer to BBClass of an Object's Type
return: Pointer to BBClass of an Object's Type
EndRem
Function ObjectClass:Int Ptr(o:Object)
	Local ref:Int Ptr = Int Ptr ObjectToPtr(o)
	Return Int Ptr ref[0]
EndFunction

Rem
bbdoc: Checks if two Objects are of the same Type
return: True if same Type
Otherwise False
EndRem
Function TypeCompare(o1:Object, o2:Object)
	Return ObjectClass(o1) = ObjectClass(o2)
EndFunction

Rem
bbdoc: ObjectEx
An #ObjectEx is a summary or a map of a Type, which is used to implement
further quite powerful features for managing them.
End Rem
Type ObjectEx
	
	Rem
	bbdoc: Name of the Type that the ObjectEx represents
	EndRem
	Field name:String
	
	Rem
	bbdoc: Pointer to the BBClass structure
	about:
	Every Type has its own BBClass structure. It saves Methods and Functions,
	a pointer to the Super class, the @debug_scope, the size of an instance
	of that class etc.
	EndRem 
	Field class:Int Ptr
	
	Rem
	bbdoc: @size represents the size of an instance of the Type the ObjectEx was
	generated for in bytes.
	EndRem
	Field size
	
	Rem
	bbdoc: @ident saves all Identifiers of the Type (Const, Field, Method,
	Function) except Globals. This is unfortunately due to the BlitzMax Compiler
	not including the Globals into the @debug_scope, so I cannot discover them.
	EndRem
	Field ident:TMap
	
	Rem
	bbdoc: @instances saves all Objects of the type of the ObjectEx that you
	saved using @add 
	EndRem
	Field instances:TMap
	
	Method New()
		ident = CreateMap()
		instances = CreateMap()	
	EndMethod
	
	Rem
	bbdoc: Creates an ObjectEx from an Object instance of any type
	returns: An ObjectEx
	about:
	This gathers information in the instance's and its @debug_scope's
	structures for creating an ObjectEx. 
	EndRem
	Function Create:ObjectEx(inst:Object)
		Local oex:ObjectEx = New ObjectEx
		oex.class = ObjectClass(inst)
		oex.size = oex.class[2]
		oex.name = String.FromCString(ParseObject(oex, inst, oex_cb))	
		Return oex
	EndFunction
		
	Rem
	bbdoc: Create ObjectEx for Super Class of existing ObjectEx
	returns: ObjectEx of Super Class
	EndRem
	Method SuperObjectEx:ObjectEx()
		Local oex:ObjectEx = New ObjectEx
		oex.class = Int Ptr(Self.class[0])
		oex.size = oex.class[2]
		oex.name = String.FromCString(ParseClass(oex, oex.class, oex_cb))	
		Return oex
	EndMethod
		
	Rem
	bbdoc: @Add saves an Object in the ObjectEx
	returns: True on success
	Otherwise False
	about:
	Saves an Object which can then be accessed by @name in the
	ObjectEx. If @overwrite is true, an existing Object with the
	same @name (case-insensitive!) is overwritten.
	EndRem
	Method Add(name:String, inst:Object, overwrite = 0)
		Local n:String
		If ObjectClass(inst) <> class Then Return False
		If Not overwrite Then
			name = name.tolower()
			For n = EachIn MapKeys(instances)
				If n.tolower() = name Then Return False
			Next
		EndIf
		MapInsert instances, name, inst
		Return True
	EndMethod

	Rem
	bbdoc: @Remove removes an Object from the ObjectEx
	returns: True on success
	False, if there's no Object @name
	EndRem	
	Method Remove(name:String)
		If Not MapContains(instances, name) Then Return False
		MapRemove instances, name
		Return True
	EndMethod
	
	Rem
	bbdoc: Get a saved Object
	returns: The Object saved as @name
	Null if there is no Object @name
	EndRem
	Method Obj:Object(name:String)
		Return MapValueForKey(instances, name)
	EndMethod

	Rem
	bbdoc: Get all saved Objects
	returns: Array of Objects that were saved
	Empty Array if no Objects were saved
	EndRem	
	Method Objs:Object[]()
		Local os:Object[]
		For Local o:Object = EachIn MapValues(instances)
			os :+ [o]
		Next
		Return os
	EndMethod

	Rem
	bbdoc: Finds an Identifier by name in the ObjectEx
	returns: #ObjectExIdent for a given Identifier in an ObjectEx
	Returns Null if there is no Identifier called @name.
	about:
	The returned #ObjectExIdent allows you to modify the variable of a type instance.
	So you can dynamically modify variables by using their names in String literals.
	(confusing?)
	EndRem
	Method Identifier:ObjectExIdent(name:String)
		Return ObjectExIdent(MapValueForKey(ident, name))
	EndMethod
	
	Rem
	bbdoc: Search all Methods
	returns: Array of #ObjectExIdent of all Methods
	Returns an empty Array if there are no Methods.
	EndRem	
	Method Methods:ObjectExIdent[]()
		Local oexi:ObjectExIdent, m:ObjectExIdent[]
		For oexi = EachIn MapValues(ident)
			If oexi.id = 6 Then m:+[oexi] 
		Next
		Return m
	EndMethod
	
	Rem
	bbdoc: Search all Functions
	returns: Array of #ObjectExIdent of all Functions
	Returns an empty Array if there are no Functions.
	EndRem	
	Method Functions:ObjectExIdent[]()
		Local oexi:ObjectExIdent, m:ObjectExIdent[]
		For oexi = EachIn MapValues(ident)
			If oexi.id = 7 Then m:+[oexi] 
		Next
		Return m
	EndMethod
	
	Rem
	bbdoc: Search all Constants
	returns: Array of #ObjectExIdent of all Constants
	Returns an empty Array if there are no Constants.
	EndRem	
	Method Constants:ObjectExIdent[]()
		Local oexi:ObjectExIdent, m:ObjectExIdent[]
		For oexi = EachIn MapValues(ident)
			If oexi.id = 1 Then m:+[oexi] 
		Next
		Return m
	EndMethod
	
	Rem
	bbdoc: Search all Fields
	returns: Array of #ObjectExIdent of all Fields
	Returns an empty Array if there are no Fields.
	EndRem		
	Method Fields:ObjectExIdent[]()
		Local oexi:ObjectExIdent, m:ObjectExIdent[]
		For oexi = EachIn MapValues(ident)
			If oexi.id = 3 Then m:+[oexi] 
		Next
		Return m
	EndMethod

	Rem
	bbdoc: Get the value of a Constant
	returns: Value of the specific Constant the #ObjectExIdent represents
	If the ObjectExIdent does not represent a Constant this function returns Null.
	EndRem
	Method GetConstValue:String(o:Object, name:String)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 1 Then Return Null
		Return String.FromCString(ObjectToPtr(o) + i.offset)
	EndMethod
		
	
	Rem
	bbdoc: Get the value of a Field
	returns: Value of the specific Field the #ObjectExIdent represents
	If the ObjectExIdent does not represent a Field this function will return Zero
	Use #GetValue for Constants.
	about:
	This function will return an Integer. By using #IntToObject you can still
	access Types, Arrays and Strings.
	For 64 bit and floating point types you will need the functions below.
	EndRem
	Method Get(o:Object, name:String)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		Return i.GetLong(o)
	EndMethod
	
	Rem
	bbdoc: Get the value of a Long Field
	returns: Value of the specific Field the #ObjectExIdent represents
	If the ObjectExIdent does not represent a Field this function will return Zero
	about:
	This function will return an Integer. By using #IntToObject you can still
	access Types, Arrays and Strings.
	For 64 bit and floating point types you will need the functions below.
	EndRem
	Method GetLong:Long(o:Object, name:String)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		Return i.GetLong(o)
	EndMethod
	
	Rem
	bbdoc: Get the value of a Float Field
	returns: Value of the specific Field the #ObjectExIdent represents
	If the ObjectExIdent does not represent a Field this function will return Zero
	about:
	This function will return an Integer. By using #IntToObject you can still
	access Types, Arrays and Strings.
	For 64 bit and floating point types you will need the functions below.
	EndRem
	Method GetFloat:Float(o:Object, name:String)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		Return i.GetFloat(o)
	EndMethod
	
	Rem
	bbdoc: Get the value of a Double Field
	returns: Value of the specific Field the #ObjectExIdent represents
	If the ObjectExIdent does not represent a Field this function will return Zero
	about:
	This function will return an Integer. By using #IntToObject you can still
	access Types, Arrays and Strings.
	For 64 bit and floating point types you will need the functions below.
	EndRem
	Method GetDouble:Double(o:Object, name:String)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		Return i.GetDouble(o)
	EndMethod	
	
	Rem
	bbdoc: Set the value of a Field
	about:
	Set the value of the Field the #ObjectExIdent represents. This function takes
	an integer value, but by using #ObjectToInt you can also store Types, Arrays
	and Strings.
	For 64 bit and floating point types you will need the functions below.
	EndRem
	Method Set(o:Object, name:String, val)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		i.Set o, val
		Return 1
	EndMethod

	Rem
	bbdoc: Set the value of a Long Field
	about:
	Set the value of the Field the #ObjectExIdent represents.
	EndRem
	Method SetLong(o:Object, name:String, val:Long)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		i.SetLong o, val
		Return 1
	EndMethod
	
	Rem
	bbdoc: Set the value of a Float Field
	about:
	Set the value of the Field the #ObjectExIdent represents.
	EndRem
	Method SetFloat(o:Object, name:String, val:Float)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		i.SetFloat o, val
		Return 1
	EndMethod
	
	Rem
	bbdoc: Set the value of a Double Field
	about:
	Set the value of the Field the #ObjectExIdent represents.
	EndRem
	Method SetDouble(o:Object, name:String, val:Double)
		Local i:ObjectExIdent = Identifier(name)
		If Not i Or i.id <> 3 Then Return 0
		i.SetDouble o, val
		Return 1
	EndMethod
		
	Method new_ident(id, name:String, typ:String, offset)
		MapInsert ident, name, ObjectExIdent.Create(id, name, typ, offset)
	EndMethod

EndType


Type ObjectExIdent

	Rem
	bbdoc: Name of Variable/Constant/Method/Function
	EndRem
	Field name:String
	Rem
	bbdoc: Type Tag. Specifies type of a Variable or return type and parameter list of Functions/Methods
	EndRem
	Field typ:String
	Field offset
	Field id
	
	Function Create:ObjectExIdent(id, name:String, typ:String, offset)
		Local oexi:ObjectExIdent = New ObjectExIdent
		oexi.name = name
		oexi.typ = typ
		oexi.offset = offset
		oexi.id = id
		Return oexi
	EndFunction

	Rem
	bbdoc: Get the value of a Field
	returns: Value of the specific Field the #ObjectExIdent represents
	EndRem
	Method Get(o:Object)
		Return (Int Ptr(ObjectToPtr(o)+offset))[0]
	EndMethod
	
	Rem
	bbdoc: Get the value of a Long Field
	returns: Value of the specific Field the #ObjectExIdent represents
	EndRem
	Method GetLong:Long(o:Object)
		Return (Long Ptr(ObjectToPtr(o)+offset))[0]
	EndMethod
	
	Rem
	bbdoc: Get the value of a Float Field
	returns: Value of the specific Field the #ObjectExIdent represents
	EndRem
	Method GetFloat:Float(o:Object)
		Return (Float Ptr(ObjectToPtr(o)+offset))[0]
	EndMethod
	
	Rem
	bbdoc: Get the value of a Double Field
	returns: Value of the specific Field the #ObjectExIdent represents
	EndRem
	Method GetDouble:Double(o:Object)
		Return (Double Ptr(ObjectToPtr(o)+offset))[0]
	EndMethod	
	
	Rem
	bbdoc: Set the value of a Field
	EndRem
	Method Set(o:Object, val)
		(Int Ptr(ObjectToPtr(o)+offset))[0] = val
	EndMethod

	Rem
	bbdoc: Set the value of a Long Field
	EndRem
	Method SetLong(o:Object, val:Long)
		(Long Ptr(ObjectToPtr(o)+offset))[0] = val
	EndMethod
	
	Rem
	bbdoc: Set the value of a Float Field
	EndRem
	Method SetFloat(o:Object, val:Float)
		(Float Ptr(ObjectToPtr(o)+offset))[0] = val
	EndMethod
	
	Rem
	bbdoc: Set the value of a Double Field
	EndRem
	Method SetDouble(o:Object, val:Double)
		(Double Ptr(ObjectToPtr(o)+offset))[0] = val
	EndMethod

EndType


